#ifndef __QT4CUPS_H
#define __QT4CUPS_H

#include <QObject>
#include <QStringList>
#include <QVariant>
#include <QDateTime>
#include <QByteArray>
#include <QTextCodec>

#include <cups/cups.h>
#include <cups/http.h>

#include <fly-admin-printer_export.h>

namespace Cups
{
extern FLY_ADMIN_PRINTER_EXPORT QTextCodec *codec;

void dumpAttributes(const QVariantMap &attrs);

////////////////////////////////////////////////////////////////////////////////
struct FLY_ADMIN_PRINTER_EXPORT Job
{
    int id;
    QString dest;
    QString title;
    QString user;
    QString format;
    ipp_jstate_t state;
    int size;
    int priority;
    QDateTime completed_time;
    QDateTime creation_time;
    QDateTime processing_time;
    Job(const cups_job_t &job)
    {
        id = job.id;
        if (job.dest) dest = codec->toUnicode(job.dest);
        if (job.title) title = codec->toUnicode(job.title);
        if (job.user) user = codec->toUnicode(job.user);
        if (job.format) format = codec->toUnicode(job.format);
        state = job.state;
        size = job.size;
        priority = job.priority;
        completed_time = QDateTime::fromTime_t(job.completed_time);
        creation_time = QDateTime::fromTime_t(job.creation_time);
        processing_time = QDateTime::fromTime_t(job.processing_time);
    }
};


////////////////////////////////////////////////////////////////////////////////

//int cupsAdminExportSamba(const QString &dest, const QString &ppd,
//    const QString &samba_server, const QString &samba_user,
//    const QString &samba_password, const QString &logfile=QString());

////////////////////////////////////////////////////////////////////////////////

QString statusName(ipp_status_t);
QString tagName(ipp_tag_t);
QString opName(ipp_op_t);
QString orientName(ipp_orient_t);

class FLY_ADMIN_PRINTER_EXPORT HTTP: public QObject
{
    Q_OBJECT
private:
    http_t *http_;
private:
    HTTP(http_t *http): QObject(), http_(http) {}

    static HTTP* create(const QString &host, int port, http_encryption_t encryption);
    static HTTP* create();

    ipp_pstate_t destState(const QString &uri);
public:
    enum JobCompletion
    {
        All = -1, Active = 0, Completed = 1
    };

    inline http_t *handle() const
    {
        return http_;
    }

    ~HTTP()
    {
        if (http_)
            ::httpClose(http_);
    }
    void reconnect();
    //constructor
    static HTTP* getUniqueConnection();

    QMap <QString, QStringList> cupsAdminGetServerSettings();

    inline bool isNull() const
    {
        return http_ == 0;
    }

    QList <QVariantMap> destinations();
    QVariantMap destAttrs(const QString &uri);
    bool setPrinterPolicy(const QString &uri, const QString &policy);
    bool setPrinterMac(const QString &uri, const QString &min, const QString &max);
    QList < QMap <QString, QString> > cupsDestinations(const QString &name = "");
    QList <Job> jobs(const QString &dest = QString::null, bool myjobs = false, JobCompletion completed = Active);
    QStringList backends();
    QList <QVariantMap> devices();
    QList <QVariantMap> PPDs();
    ipp_pstate_t printerState(const QString &name);
    ipp_pstate_t classState(const QString &name);
    bool enablePrinter(const QString &uri);
    bool disablePrinter(const QString &uri);
    bool printerSetServerDefault(const QString &uri);
    bool addPrinter(const QString &uri, const QString &name, const QString &ppd, const QString &location = QString::null, const QString &description = QString::null, bool shared = false);
    bool addClass(const QString &name, const QStringList &uris, const QString &location = QString::null, const QString &description = QString::null, bool shared = false);
    bool deleteClass(const QString &uri);
    bool deletePrinter(const QString &uri);
    bool holdJob(const QString &uri, int id);
    bool releaseJob(const QString &uri, int id);
    bool cancelJob(const QString &dest, int id);
    bool restartJob(const QString &dest, int id);
signals:
    void step(int steps);
};


////////////////////////////////////////////////////////////////////////////////
class FLY_ADMIN_PRINTER_EXPORT IPP : public QObject
{

    Q_OBJECT
    ipp_t *ipp_;

public:

    class Value
    {
    private:
        ipp_value_t value_;
    public:
        Value(const ipp_value_t &value): value_(value) {};

        struct Resolution
        {
            int width, height;
            ipp_res_t units;

            Resolution()
            {
                width = height = 0;
                units = IPP_RES_PER_INCH;
            }
            Resolution(int w, int h, ipp_res_t u)
            {
                width = w;
                height = h;
                units = u;
            }
            Resolution(const QVariant &src)
            {
                QStringList l = src.toString().split(":");
                if (l.count() == 3)
                {
                    width = l[0].toInt();
                    height = l[1].toInt();
                    units = (ipp_res_t)l[2].toInt();
                }
            }
            operator QVariant() const
            {
                return QString("%1:%2:%3").arg(width).arg(height).arg(units);
            }
        };
        struct Range
        {
            int min, max;
            Range()
            {
                min = max = 0;
            }
            Range(int min, int max): min(min), max(max) { }
            Range(const QVariant &src)
            {
                QStringList l = src.toString().split(":");
                if (l.count() == 2)
                {
                    min = l[0].toInt();
                    max = l[1].toInt();
                }
            }
            bool isNull() const
            {
                return (min == 0 && max == 0);
            }
            operator QVariant() const
            {
                return QString("%1:%2").arg(min).arg(max);
            }
        };
        inline int asInt() const
        {
            return value_.integer;
        }
        inline bool asBool() const
        {
            return value_.boolean;
        }
        inline QDateTime asDateTime() const
        {
            return QDateTime::fromTime_t(::ippDateToTime(value_.date));
        }
        inline Resolution asResolution() const
        {
            return Resolution(value_.resolution.xres, value_.resolution.yres, value_.resolution.units);
        }
        inline Range asRange() const
        {
            return Range(value_.range.lower, value_.range.upper);
        }
        QString asString() const;
        inline QByteArray asData() const
        {
            return QByteArray((const char*)value_.unknown.data, value_.unknown.length);
        }
        QVariant convert(ipp_tag_t tag) const;
    };

    enum PrinterType
    {
        Class = 0x00000001,     //Is a printer class.
        Remote = 0x00000002,        //Is a remote destination.
        Black = 0x00000004,     //Can print in black.
        Color = 0x00000008,     //Can print in color.
        Double = 0x00000010,        //Can print on both sides of the page in hardware.
        Staple = 0x00000020,        //Can staple output.
        Copy = 0x00000040,      //Can do fast copies in hardware.
        Collation = 0x00000080,     //Can do fast copy collation in hardware.
        Punch = 0x00000100,     //Can punch output.
        Cover = 0x00000200,     //Can cover output.
        Bind = 0x00000400,      //Can bind output.
        Sort = 0x00000800,      //Can sort output.
        Small = 0x00001000,     //Can handle media up to US-Legal/A4.
        Medium = 0x00002000,        //Can handle media from US-Legal/A4 to ISO-C/A2.
        Large = 0x00004000,     //Can handle media larger than ISO-C/A2.
        Custom = 0x00008000,        //Can handle user-defined media sizes.
        Implicit = 0x00010000,      //Is an implicit (server-generated) class.
        Default = 0x00020000,       //Is the a default printer on the network.
        Fax = 0x00040000,           //Is a facsimile device.
        Reject = 0x00080000,        //Is rejecting jobs.
        Delete = 0x00100000,        //Delete this queue.
        NotShared = 0x00200000,     //Queue is not shared.
        ReqAuth = 0x00400000,       //Queue requires authentication.
        Command = 0x00800000,       //Queue supports CUPS command files.
        AutoDiscovery = 0x01000000  //Queue was automatically discovered and added.
    };

    Q_DECLARE_FLAGS (PrinterTypeFlags, PrinterType)

    static QString typeName(PrinterType type);
    static IPP* create(ipp_op_t op);
    static IPP* create();

    ~IPP();

    inline ipp_t *handle()
    {
        return ipp_;
    }

    inline bool isNull() const
    {
        return ipp_ == 0;
    }

    inline ipp_state_t state() const
    {
        return ( ipp_ )
               ? ipp_->state
               : IPP_ERROR;
    }
    inline QString stateName() const
    {

        if ( ! ipp_ || ipp_->state == IPP_ERROR ) return "IPP_ERROR";
        if (ipp_->state == IPP_IDLE) return "IPP_IDLE";
        if (ipp_->state == IPP_HEADER) return "IPP_HEADER";
        if (ipp_->state == IPP_ATTRIBUTE) return "IPP_ATTRIBUTE";
        if (ipp_->state == IPP_DATA) return "IPP_DATA";
        else return "";
    }
    inline void setState(ipp_state_t state)
    {
        if ( ipp_)
        {
            ipp_->state = state;
        }
    }

    inline ipp_request_t& request()
    {
        Q_ASSERT( ipp_ );
        return ipp_->request;
    }


    const ipp_attribute_t *firstAttribute() const
    {
        Q_ASSERT( ipp_ );
        return ipp_->attrs;
    }

    const ipp_attribute_t *lastAttribute() const
    {
        Q_ASSERT( ipp_ );
        return ipp_->last;
    }

    unsigned attributeCount() const
    {
        if ( ! ipp_ )
        {
            return 0;
        }
        unsigned counter = 0;
        for (const ipp_attribute_t *i = ipp_->attrs; i && i != ipp_->last; i = i->next)
        {
            counter++;
        }
        return counter;
    }

    QList <QVariantMap> attributes() const;
    QMap <QString, ipp_tag_t> valueTags() const;
    QMap <QString, ipp_tag_t> groupTags() const;


    void addAttribute(const QString &name, ipp_tag_t tag, const QStringList &values);

    inline void addAttribute(const QString &name, ipp_tag_t tag, const QString &value)
    {
       if( ipp_ )
       {
           ::ippAddString(ipp_, IPP_TAG_OPERATION, tag, codec->fromUnicode(name).data(), NULL, codec->fromUnicode(value).data());
       }
    }
    inline void addPrinterAttribute(const QString &name, ipp_tag_t tag, const QString &value)
    {
        qDebug("addPrinterAttribute(name=%s, typeTag=%d, value=%s",
               qPrintable(name), tag, qPrintable(value));
        if( ipp_ )
        {
            ::ippAddString(ipp_, IPP_TAG_PRINTER, tag,
                           codec->fromUnicode(name).data(), NULL,
                           codec->fromUnicode(value).data());
        }
    }

    inline void addAttribute(const QString &name, int value)
    {
        if( ipp_ )
        {
            ::ippAddInteger(ipp_, IPP_TAG_OPERATION, IPP_TAG_INTEGER, codec->fromUnicode(name).data(), value);
        }
    }

    inline void addAttribute(ipp_tag_t tag, const QString &name, int value)
    {
        if( ipp_ )
        {
            ::ippAddInteger(ipp_, IPP_TAG_OPERATION, tag, codec->fromUnicode(name).data(), value);
        }
    }

    inline void addAttribute(const QString &name, bool value)
    {
        if( ipp_ )
        {
            ::ippAddBoolean(ipp_, IPP_TAG_OPERATION, codec->fromUnicode(name).data(), value);
        }
    }

    void setOperation(ipp_op_t op)
    {
        if( ipp_ )
        {
            ipp_->request.op.operation_id = op;
        }
    }

    IPP *doRequest(HTTP *connection, const QString &resource);
    IPP *doRequest(HTTP *connection, const QString &resource, const QString &filename);

protected:

    IPP(ipp_t * /*ipp*/ );

};

class FLY_ADMIN_PRINTER_EXPORT PPD: public QObject
{
    Q_OBJECT
    void init(const QString &filename);
public:
    QVariantMap contents;

    PPD(HTTP *http, const QString &printer, QObject *parent = 0);
    PPD(const QString &filename, QObject *parent = 0);

};

} //namespace

#endif//__QT4CUPS_H
